from py2neo import Graph
import getpass
from collections import defaultdict


class Neo:
    def __init__(self, uri, user, password=None):
        if password is None:
            password = getpass.getpass("Password:")
        self.graph = Graph(uri, auth=(user, password))


    def execute(self, query):
        r = self.graph.run(query)
        return r
    def bucket_concepts(self, data, bucket_size_seconds):
        entities = data['entities']

        _bucket = []
        _concepts = set()
        start_time = -1
        new_stream = []
        for ent in entities:
            if start_time == -1:
                start_time = ent['timestamp']

            if ent['timestamp'] - start_time >= bucket_size_seconds:
                # Add to stream
                new_stream.extend(_bucket)
                _bucket = []
                _concents = set()
                start_time = ent['timestamp']

                new_stream.append({'name': '<SEP>'})

            if ent['conceptId'] not in _concepts:
                _bucket.append(ent)
                _concepts.add(ent['conceptId'])

        if _bucket:
            new_stream.extend(_bucket)

        data['entities'] = new_stream

    def get_all_patients(self, concepts, limit=1000):
        r''' Return all patients having all concepts
        '''

        q = "WITH [{}] AS cs ".format(",".join(["'{}'".format(c) for c in concepts]))
        q += '''MATCH (c:Concept)--(d:Document)--(pt:Patient)
        WHERE c.conceptId in cs
        WITH pt, size(cs) as inputCnt, count(DISTINCT c) as cnt
        WHERE cnt = inputCnt
        RETURN pt LIMIT 20
        '''
        data = self.execute(q).data() # Do not like this too much 

        return [n['pt']['patientId'] for n in data]


    def get_all_concepts_from(self, patient_id=None, document_id=None,
            concept_type=None, limit=1000, bucket_size_seconds=None,
            min_count=0, meta_requirements=None):
        r''' Returns all concepts belonging to a document or patient
        given the concept type (if none all are retruned).
        '''

        q_concept = '(concept:Concept'
        if concept_type is not None:
            q_concept += ' {type: %s} ' % concept_type
        else:
            q_concept += ') '

        if patient_id is not None:
            q = 'MATCH (patient:Patient {patientId: "%s"})--(document:Document)-[has:HAS]-' % patient_id
        else:
            q = 'MATCH (patient:Patient)--(document:Document {documentId: "%s"})-[has:HAS]-' % document_id
        q += q_concept
        q += 'RETURN patient, document, concept, has LIMIT %s' % limit

        data = self.execute(q).data() # Do not like this too much 
        out = None
        if len(data) > 0:
            out = {'patient': data[0]['patient']['patientId'],
                   'entities': []}

            ent = {}
            cnt = defaultdict(int)
            for row in data:
                if meta_requirements is None or all([row['has'][meta] == value for meta,value in meta_requirements.items()]):
                    ent['conceptId'] = row['concept']['conceptId']
                    ent['name'] = row['concept']['name']
                    ent['documentId'] = row['document']['documentId']
                    ent.update(row['has']) # add all the stuff from the meta ann

                    out['entities'].append(ent)
                    cnt[ent['conceptId']] += 1
                    ent = {}

            # Cleanup based on min_count
            new_ents = []
            for ent in out['entities']:
                if cnt[ent['conceptId']] >= min_count:
                    ent['count'] = cnt[ent['conceptId']]
                    new_ents.append(ent)
            out['entities'] = new_ents

            if bucket_size_seconds is not None:
                self.bucket_concepts(data=out, bucket_size_seconds=bucket_size_seconds)

        return out
